@@ -42,15 +42,26 @@
  */
 #define DEVICE_FLAG_NONE 0x00000000
 /**
- * This means the device supports both MTP and USB mass 
- * storage by dynamically reconfiguring itself if it is not
- * used with MTP before a certain timeout.
+ * This means that the PTP_OC_MTP_GetObjPropList is broken
+ * in the sense that it won't return properly formatted metadata
+ * for ALL files on the device when you request an object 
+ * property list for object 0xFFFFFFFF with parameter 3 likewise
+ * set to 0xFFFFFFFF. Compare to 
+ * DEVICE_FLAG_BROKEN_MTPGETOBJECTPROPLIST which only signify
+ * that it's broken when getting metadata for a SINGLE object.
  */
-#define DEVICE_FLAG_DUALMODE 0x00000001
+#define DEVICE_FLAG_BROKEN_MTPGETOBJPROPLIST_ALL 0x00000001
 /**
  * This means that under Linux, another kernel module may 
- * be using the MTP interface, so we need to detach it if
- * it is.
+ * be using this device's USB interface, so we need to detach 
+ * it if it is. Typically this is on dual-mode devices that
+ * will present both an MTP compliant interface and device
+ * descriptor *and* a USB mass storage interface. If the USB
+ * mass storage interface is in use, other apps (like our
+ * userspace libmtp through libusb access path) cannot get in
+ * and get cosy with it. So we can remove the offending 
+ * application. Typically this means you have to run the program
+ * as root as well.
  */
 #define DEVICE_FLAG_UNLOAD_DRIVER 0x00000002
 /**
@@ -75,6 +86,18 @@
  * 512 bytes, it will work with USB 1.1 and USB 2.0 alike.
  */
 #define DEVICE_FLAG_NO_ZERO_READS 0x00000008
+/**
+ * This flag means that the device is prone to forgetting the
+ * OGG container file type, so that libmtp must look at the
+ * filename extensions in order to determine that a file is
+ * actually OGG. This is a clear and present firmware bug, and
+ * while firmware bugs should be fixed in firmware, we like
+ * OGG so much that we back it by introducing this flag.
+ * The error has only been seen on iriver devices. Turning this
+ * flag on won't hurt anything, just that the check against
+ * filename extension will be done for files of "unknown" type.
+ */
+#define DEVICE_FLAG_IRIVER_OGG_ALZHEIMER 0x00000010
 
 /**
  * Internal USB struct.
